"""Main CLI g-storm interface.

Enables building static-typed ORM models.
"""

import click
from . import scripts

@click.group()
def cli():
  pass

def main():
  cli.add_command(scripts.cleanup)
  cli.add_command(scripts.validate_schema)
  cli.add_command(scripts.build_classes)
  # cli.add_command(scripts.build_query)
  # cli.add_command(scripts.build_create)
  # cli.add_command(scripts.build_upsert)
  cli()

if __name__ == "__main__":
  main()


import pydash as __ 
from typing import Optional, List, _GenericAlias
from datetime import datetime
from gstorm.helpers.gql import setup_gql
from pprint import pprint
import attr
from dataclasses import dataclass, field
from gstorm.QueryBuilder import QueryBuilder
from gstorm.enums import QueryType
from gstorm.BaseGraphQLType import BaseGraphQLType
from gstorm.helpers.date_helpers import get_local_date, iso8601_to_local_date

def nint(value):
  if value is None:
    return None
  elif type(value) == str:
    return int(value)
  elif type(value) == int:
    return value
  else:
    raise Exception(f"Cannot cast {type(value)} to nullable int")

@attr.s
class GraphQLType(BaseGraphQLType):
  '''Base class for Graphql types,
  has one python's dataclass field per graphql type.
  Plus a __type field with the Type's metadata (opt-out, etc)
  '''
  # __metadata__: str = attr.ib(factory=dict) # attr.ib(default={'opt-out': {}}) # ! MAY BE REQUIRED LATER FOR @opt-out, etc
  __sync__: bool = attr.ib(repr=False, default=False)
  __errors__: List[dict] = attr.ib(repr=False, factory=list)
  id: Optional[int] = attr.ib(default=None, converter=nint, metadata={'unique': {}})
  insertedAt: datetime = attr.ib(repr=False, factory=lambda: get_local_date(datetime.now()), converter=iso8601_to_local_date, metadata={'readonly': True})
  updatedAt: datetime = attr.ib(repr=False, factory=lambda: get_local_date(datetime.now()), converter=iso8601_to_local_date, metadata={'readonly': True})
  
  def __getitem__(self, name):
    return self.__getattribute__(name)
  @classmethod
  def get_field(cls, name):
    return attr.fields_dict(cls).get(name)
  @classmethod
  def get_fields(cls):
    return attr.fields_dict(cls)
  @classmethod
  def get_base_fields(cls):
    list_fields = [
      name
      for name, field in attr.fields_dict(cls).items()
      if not field.type == _GenericAlias # if not type List
        and not GraphQLType in (
        __.get(field, 'type.__mro__')
        or __.get(field, 'type.__dict__.__args__')
        )
        and not name.startswith('_')
    ]
    return list_fields
  @classmethod
  def get_list_fields(cls):
    list_fields = [
      name
      for name, field in attr.fields_dict(cls).items()
      if field.type == _GenericAlias # if not type List
    ]
    return list_fields
  @classmethod
  def get_object_fields(cls):
    obj_fields = [
      name
      for name, field in attr.fields_dict(cls).items()
      if GraphQLType in (
        __.get(field, 'type.__mro__')
        or __.get(field, 'type.__dict__.__args__')
      )
    ]
    return obj_fields
  @classmethod
  def get_unique_fields(cls):
    return [
      name
      for name, field in attr.fields_dict(cls).items()
      if 'unique' in field.metadata
    ]
  @classmethod
  def get_metadata(cls, field_name):
    field = attr.fields_dict(cls)[field_name]
    return field.metadata
  def create(self):
    pass
  def load(self):
    raise NotImplementedError('Child class should implement load method')
  @classmethod
  def query(cls):
    return QueryBuilder(cls)
  @classmethod
  def query_one(cls):
    return QueryBuilder(cls, QueryType.SINGULAR)
  def is_sync(self):
    return self.__sync__
  def has_errors(self):
    return len(self.__errors__) > 0
  def get_errors(self):
    return self.__errors__
  # ! method aliases
  # query
  q = query
  qm = query
  qp = query # query plural
  # query_one
  q1 = query_one
  qs = query_one # query singular



from typing import Dict, List
from dataclasses import dataclass, field
from datetime import datetime
import inflect
from pygqlc import GraphQLClient
from gstorm.BaseGraphQLType import BaseGraphQLType
from gstorm.enums import QueryType
from gstorm.helpers.str_handling import remove_capital, contains_digits
from gstorm.helpers.date_helpers import get_utc_date, get_iso8601_str
from gstorm.helpers.gql import setup_gql

'''

class C(object):
    def __init__(self):
        self._x = None
    @property
    def x(self):
        """I'm the 'x' property."""
        print('Hippity hoppity, someone looked at my property!')
        return self._x
    @x.setter
    def x(self, value):
        print('setting x value!')
        self._x = value
    @x.deleter
    def x(self):
        print('deleted x value!')
        del self._x

CREATE, UPDATE, UPSERT, DELETE
# SIMPLE
mutation {
  createTank(
    name: 'L101'
    type: 'Gobierno'
    capacity: 1000
  ){
    successful
    messages{
      field
      message
    }
    result{
      id
      name
      type
      capacity
      insertedAt
      updatedAt
    }
  }
}
# PARAMS
query {
  tanks (
    filter: {
      capacity: 10
      name: "hola"
      after: {
        attribute: INSERTED_AT
        date: "2020-01-01T00:00:00Z"
      }
    }
    orderBy: {
      desc: ID
    }
    limit: 10
  ){
    id
    name
    capacity
    insertedAt
    updatedAt
  }
}
'''

class MutationBuilder():
  pass


import os
import csv
from datetime import datetime
from colorama import Fore, Back, Style
import sys
from ValiotWorker.Logging import (log, LogLevel)

def silent_removefile(filename):
  try:
    os.remove(filename)
  except OSError:
    pass

def query_yes_no(question, default="yes", level=LogLevel.INFO):
  """Ask a yes/no question via raw_input() and return their answer.

  "question" is a string that is presented to the user.
  "default" is the presumed answer if the user just hits <Enter>.
      It must be "yes" (the default), "no" or None (meaning
      an answer is required of the user).

  The "answer" return value is True for "yes" or False for "no".
  """
  valid = {
    "yes": True,
    "y": True,
    "ye": True,
    "no": False,
    "n": False
  }
  if default is None:
    prompt = " [y/n] "
  elif default == "yes":
    prompt = " [Y/n] "
  elif default == "no":
    prompt = " [y/N] "
  else:
    raise ValueError("invalid default answer: '%s'" % default)

  while True:
    log(level, question + prompt)
    choice = input().lower()
    if default is not None and choice == '':
      return valid[default]
    elif choice in valid:
      return valid[choice]
    else:
      log(LogLevel.WARNING, "Please respond with (yes/no) or (y/n).")



import os
from ValiotWorker.Logging import log, LogLevel

def inmemory_cleanup(src):
  src_path = os.path.join(os.getcwd(), src)
  output = []
  with open(src_path, 'r') as src_file:
    for line in src_file:
      is_whitespace = not line.strip()
      if is_whitespace:
        output.append(line)
        continue
      line = line.rstrip() # clean
      line = line.split('#')[0].rstrip() # remove comment
      # append if not empty
      if line:
        output.append(line + '\n')
    output.append('\n')
  return output


def scalar_map(field_typename):
  return {
    'Boolean': 'Boolean',
    'Integer': 'Int',
    'Float': 'Float',
    'String': 'String',
    'Text': 'Text',
    'Json': 'map',
    'Date': 'Date',
    'Time': 'Time',
    'Datetime': 'DateTime'
  }[field_typename]

def get_field_kind(field_typename):
  scalar_types = [
    'Boolean',
    'Integer',
    'Float',
    'String',
    'Text',
    'Json',
    'Date',
    'Time',
    'Datetime'
  ]
  if field_typename in scalar_types:
    return 'SCALAR'
  else:
    return 'OBJECT'

'''
Syntactic analizer for a graphql schema

things to consider:

# ! 1. general structure of a type
type TYPE_NAME @opts {
  ...
}
# ! 2. general structure of an enum
enum ENUM_TYPE @opts {
  ...
}
# ! 3. All basic types
type MyType {
  enabled: Boolean
  quantity: Int
  weight: Float
  name: String
  description: Text
  config: Json
  shipAt: Datetime
}
# ! 4. Complex type relations
# ? 4.1 ComplexType1.ListOf.ComplexType2 (Cannot be a list of BasicType)
# ? 4.2 ComplexType1.hasOne.ComplexType2 (Cannot be a hasOne of BasicType)
# ? 4.3 ComplexType1.ComplexType2 (ComplexType2 Must exist)
type User {
  name: String!
  address: Address @has_one
  posts: [Blogpost]
}

type Address {
  street: String
  city: String
  number: Integer
}

type Blogpost {
  author: User
  title: String @unique
}

type UserSkill {
  user: User
  skill: Skill
  level: Integer!
}

type Skill {
  name: String
}
'''
def gql_schema_parse(schema_file_lines):
  state = 'TYPE_SEARCH'
  schema = {} # dict of typenames
  # context
  current_type = None
  current_field = None
  for line in schema_file_lines:
    line = line.strip()
    tokens = line.split()
    if (state == 'TYPE_SEARCH'):
      if len(tokens) == 0:
        continue
      if tokens[0] == 'type':
        state = 'PARSING_TYPE'
        current_type = {}
        current_type['name'] = tokens[1]
        current_type['kind'] = 'TYPE'
        current_type['fields'] = []
      elif tokens[0] == 'enum':
        state = 'PARSING_ENUM'
        current_type = {}
        current_type['name'] = tokens[1]
        current_type['kind'] = 'ENUM'
        current_type['enumValues'] = []
      else:
        continue
    elif (state == 'PARSING_TYPE'):
      if len(tokens) == 0:
        # blank line inside the type
        continue
      elif tokens[0] == '}':
        # end of type definition
        state = 'TYPE_SEARCH'
        schema[current_type['name']] = current_type
        current_type = None
      else:
        # type body (should be a field definition)
        field_name = tokens[0].strip(':')
        field_is_required = '!' in tokens[1]
        field_is_list = '[' in tokens[1] and ']' in tokens[1]
        field_typename = tokens[1].strip('!').strip('[]').strip('!')
        field_kind = get_field_kind(field_typename)
        field_has_one = False if len(tokens) < 3 else '@has_one' == tokens[2]
        field_is_unique = False if len(tokens) < 3 else '@unique' == tokens[2]
        current_field = {}
        # current_field['required'] = field_is_required
        # current_field['is_unique'] = field_is_unique
        current_field['unique'] = field_is_unique
        current_field['has_one'] = field_has_one
        current_field['name'] = field_name
        current_field['type'] = {
          'name': field_typename if field_kind == 'OBJECT' else scalar_map(field_typename),
          'kind': 'LIST' if field_is_list else field_kind
        }
        current_type['fields'].append(current_field)
        current_field = None
    elif (state == 'PARSING_ENUM'):
      if len(tokens) == 0:
        # blank line inside the enum
        continue
      elif tokens[0] == '}':
        # end of enum definition
        state = 'TYPE_SEARCH'
        schema[current_type['name']] = current_type
        current_type = None
      else:
        # enum body (should be a EnumValue definition)
        current_type['enumValues'].append(tokens[0].strip())
  # we now know every schema type, we need to update the fields type kind to ENUM if needed:
  for typename, item in schema.items():
    if item['kind'] == 'ENUM':
      continue
    for index, field in enumerate(item['fields']):
      if not field['type']['name'] in schema:
        # may be an error, but leave it to the validation function
        continue
      if (
        field['type']['kind'] == 'OBJECT'
        and schema[field['type']['name']]['kind'] == 'ENUM'
      ):
        schema[typename]['fields'][index]['type']['kind'] = 'ENUM'
  return schema

def load_schema_from_file(src):
  try:
    clean_schema_lines = inmemory_cleanup(src)
  except FileNotFoundError as e:
    log(LogLevel.ERROR, f'File {src} not found')
  return gql_schema_parse(clean_schema_lines)


import os
from pprint import pformat
import click
from ValiotWorker.Logging import log, LogLevel
from gstorm.helpers.gql_schema_helpers import load_schema_from_file

def gql_schema_validate(schema):
  errors = []
  for typename, item in schema.items():
    log(LogLevel.INFO, f'\nvalidating type {typename}')
    # ! Style validation
    if typename[0].islower() or '_' in typename:
      # * No lowercasing Types
      errors.append({
        'type': item['name'],
        'field': field['name'],
        'message': f'Typename "{typename}" should be in CapitalCase'
      })
    if item['kind'] == 'TYPE':
      # ! validate every field
      for field in item['fields']:
        log(LogLevel.INFO, f'\tvalidating field {field["name"]}')
        # ! Style validation
        if field['name'][0].isupper():
          # * NO Capitalizing fields
          errors.append({
            'type': item['name'],
            'field': field['name'],
            'message': f'Field "{field["name"]}" should not be capitalized'
          })
        elif '_' in field['name']:
          errors.append({
            'type': item['name'],
            'field': field['name'],
            'message': f'Field "{field["name"]}" should be in camelCase'
          })
        # ! Schema validation
        if field['type']['kind'] == 'SCALAR':
          # u all right, fam
          continue
        elif field['type']['kind'] == 'LIST':
          # first of all, this type exists?
          if not field['type']['name'] in schema:
            errors.append({
              'type': item['name'],
              'field': field['name'],
              'message': f'LIST type "{field["type"]["name"]}" does not exist'
            })
            continue
          # validate List Referring type has a field with corresponding type
          referred_type = schema[field['type']['name']]
          referred_fields = [
            ref_field for ref_field in referred_type['fields']
            if ref_field['type']['name'] == typename
            and not ref_field['has_one']
          ]
          if len(referred_fields) == 0:
            errors.append({
              'type': item['name'],
              'field': field['name'],
              'message': 'LIST Has no corresponding object in referred type'
            })
          # elif len(referred_fields) > 1:
          #   errors.append({
          #     'type': item['name'],
          #     'field': field['name'],
          #     'message': 'LIST has too many referred types (Should be only one)'
          #   })
        elif field['type']['kind'] == 'OBJECT' and field['has_one']:
          # first of all, this type exists?
          if not field['type']['name'] in schema:
            errors.append({
              'type': item['name'],
              'field': field['name'],
              'message': f'OBJECT type "{field["type"]["name"]}" does not exist'
            })
            continue
          # validate Object Referring type has a field with corresponding type
          referred_type = schema[field['type']['name']]
          referred_fields = [
            ref_field for ref_field in referred_type['fields']
            if ref_field['type']['name'] == typename
            and not ref_field['has_one']
          ]
          if len(referred_fields) == 0:
            errors.append({
              'type': item['name'],
              'field': field['name'],
              'message': 'HAS_ONE has no corresponding object in referred type'
            })
          # elif len(referred_fields) > 1:
          #   errors.append({
          #     'type': item['name'],
          #     'field': field['name'],
          #     'message': 'HAS_ONE has too many referred types (Should be only one)'
          #   })
        elif field['type']['kind'] == 'OBJECT' and not field['has_one']:
          # first of all, this type exists?
          if not field['type']['name'] in schema:
            errors.append({
              'type': item['name'],
              'field': field['name'],
              'message': f'Object field type "{field["type"]["name"]}" does not exist'
            })
            continue
          # validate List Referring type has a field with corresponding type
          referred_type = schema[field['type']['name']]
          referred_fields_has_one = [
            ref_field for ref_field in referred_type['fields']
            if ref_field['type']['name'] == typename
            and ref_field['has_one']
          ]
          referred_fields_list = [
            ref_field for ref_field in referred_type['fields']
            if ref_field['type']['name'] == typename
            and ref_field['type']['kind'] == 'LIST'
          ]
          if (
            len(referred_fields_has_one) > 0
            and len(referred_fields_list) > 0
          ):
            errors.append({
              'type': item['name'],
              'field': field['name'],
              'message': 'OBJECT reference is both HAS_ONE and LIST in referred type'
            })
          elif (
            len(referred_fields_has_one) == 0
            and len(referred_fields_list) == 0
          ):
            errors.append({
              'type': item['name'],
              'field': field['name'],
              'message': f'field {field["name"]} has no corresponding LIST or HAS_ONE'
            })
    elif item['kind'] == 'ENUM':
      # ! validate Enumeration usage
      schema_object_types = {tn2: i2 for tn2, i2 in schema.items() if i2['kind'] == 'TYPE'}
      referring_types = [
        1 for typename2, item2 in schema_object_types.items()
        if len([
          1 for field in item2['fields']
          if field['type']['name'] == item['name']
        ]) > 0
      ]
      if len(referring_types) == 0:
        errors.append({
              'type': item['name'],
              'field': '',
              'message': 'ENUM does not have any referring type (should be at least one)'
            })
  return errors

@click.command()
@click.option('--src', default=None, help='Graphql schema file to analyze')
def validate_schema(src: str):
  """Validates graphql schema relations

  Parameters
  ----------
  src : str
    relative path to schema file (.graphql)
    ex: "gstorm-cli validate-schema --src=graphql/doc/schema.graphql"
  """
  # ! THIS
  schema = load_schema_from_file(src)
  from pprint import pprint
  pprint(schema)
  # ! UNTIL HERE
  validation_errors = gql_schema_validate(schema)
  if len(validation_errors) > 0:
    log(LogLevel.ERROR, f'\n\n{len(validation_errors)} ERROR FOUND INSPECTING THE SCHEMA:')
    log(LogLevel.ERROR, pformat(validation_errors))
    log(LogLevel.ERROR, '\n\n')
  else:
    log(LogLevel.SUCCESS, '\n\nNo errors found in the schema :)\n\n')



import os
import click
from .utils import one_to_one_map, imports, example
from gstorm.helpers.gql_schema_helpers import load_schema_from_file
from ValiotWorker.Logging import log, LogLevel

def _build_classes(schema, dir_to_save, mpath):

    for class_name, attributes in schema.items():
        path_and_file = os.path.join(dir_to_save, class_name + ".py")
        log(LogLevel.WARNING, f'Building class file for {class_name} model')
        with open(path_and_file, 'w') as file:

            # * Clases GraphQLType
            if attributes['kind'] == 'TYPE':
                file.write( imports ) 
                buffer_written = f"\nmpath = '{mpath}'\n\n@attr.s\nclass {class_name}(GraphQLType):\n"

                for field in attributes['fields']:
                    name = field['name']
                    kind = field['type']['kind']
                    temp = field['type']['name']
                    has_one = field['has_one']
                    unique = field['unique']
                    if any([unique, has_one]):
                        metadata = ", metadata={{{0}{1}{2}}}"\
                        .format("'unique':{}" if unique else '', \
                            ',' if all([has_one, unique]) else '', \
                            "'has_one':True" if has_one else '')
                    else: metadata = ''
                    type_name, default = (temp,"None") if temp not in one_to_one_map else one_to_one_map[temp]
                            
                    if kind == "OBJECT":
                        buffer_written += \
                            f"    {name}: GraphQLType = attr.ib(default={default}, repr=gql_repr, converter=convert_to(mpath,'{type_name}') {metadata})\n"
                    elif kind == "LIST":
                        buffer_written += \
                            f"    {name}: List[GraphQLType] = attr.ib(factory=list, repr=gql_list_repr, converter=list_convert_to(mpath,'{type_name}'), metadata={{'type':list}})\n"
                    elif kind == "ENUM":
                        file.write(f"from .{type_name} import {type_name}\n")
                        buffer_written += \
                            f"    {name}: {type_name} = attr.ib(default='{schema[type_name]['enumValues'][0]}', converter=convert_to(mpath,'{type_name}'){metadata})\n"
                    else:
                        buffer_written += \
                            f"    {name}: {type_name} = attr.ib(default={default}{metadata})\n"

            # * Clases Enum
            elif attributes['kind'] == 'ENUM':
                file.write("from enum import Enum")
                buffer_written = f"\n\nclass {class_name}(Enum):\n"     
                for enum in attributes['enumValues']:
                    buffer_written += f"    {enum} = '{enum}'\n"
            
            file.write(buffer_written)

    # Creacion del archivo __init__.py
    path_and_file = os.path.join(dir_to_save, "__init__.py")
    with open(path_and_file, 'w') as file:
        for class_name in schema:
            file.write(f"from .{class_name} import {class_name}\n")
            

@click.command()
@click.option('--src', default=None, help='Graphql schema file to analyze')
@click.option('--output', default=None, help='output directory for the GQL model files')
def build_classes(src, output):
    mpath = output.replace('/', '.')
    output = os.path.join(os.getcwd(), output)
    schema = load_schema_from_file(src)
    _build_classes(schema, output, mpath)
    log(LogLevel.SUCCESS, 'Done')



import os
from pygqlc import GraphQLClient

def setup_gql(env_name=None):
  env = env_name or os.environ.get('ENV')
  gql = GraphQLClient()
  gql.addEnvironment(
    'dev',
    url=os.environ.get('API_DEV'),
    wss=os.environ.get('WSS_DEV'),
    headers={'Authorization': os.environ.get('TOKEN_DEV')})
  gql.addEnvironment(
    'prod',
    url=os.environ.get('API'),
    wss=os.environ.get('WSS'),
    headers={'Authorization': os.environ.get('TOKEN')})
  # ! Sets the environment selected in the .env file
  gql.setEnvironment(env)
  return gql



from typing import Dict, List
from dataclasses import dataclass, field
from datetime import datetime
import inflect
from pygqlc import GraphQLClient
from gstorm.BaseGraphQLType import BaseGraphQLType
from gstorm.enums import QueryType
from gstorm.helpers.str_handling import remove_capital, contains_digits
from gstorm.helpers.date_helpers import get_utc_date, get_iso8601_str
from gstorm.helpers.gql import setup_gql

'''
# SIMPLEST
query {
  tanks {
    id
    insertedAt
    updatedAt
  }
}
# PARAMS
query {
  tanks (
    filter: {
      capacity: 10
      name: "hola"
      after: {
        attribute: INSERTED_AT
        date: "2020-01-01T00:00:00Z"
      }
    }
    orderBy: {
      desc: ID
    }
    limit: 10
  ){
    id
    name
    capacity
    insertedAt
    updatedAt
  }
}
'''

@dataclass
class QueryBuilder():
  _kind: BaseGraphQLType = None
  _type: QueryType = QueryType.PLURAL
  _filter: Dict = field(default_factory=dict)
  _findBy: Dict = field(default_factory=dict)
  _orderBy: Dict = field(default_factory=dict)
  _limit: int = 0
  _offset: int = 0
  _page: int = 0
  _query: str = ''
  _children: dict = field(default_factory=dict)
  _isChild: bool = False
  _gql: GraphQLClient = field(default_factory=setup_gql)
  def filter(self, **kwargs):
    self._filter = {**self._filter, **kwargs}
    return self
  def findBy(self, **kwargs):
    self._findBy = kwargs
    return self
  def orderBy(self, **kwargs):
    self._orderBy = kwargs
    return self
  def limit(self, count):
    self._limit = count
    return self
  def offset(self, count):
    self._offset = count
    return self
  def children(self, *args, **kwargs):
    """Appends child objects to query, accepts inputs in three formats:
      child(room=10, inventories=20)
      child({'room':10, 'inventories':20})
      child(**{'room':10, 'inventories':20})
    
    Returns
    -------
    QueryBuilder
      Same instance that called this method for chaining
    """
    if len(args) > 0:
      self._children = {**self._children, **args[0]}
    elif kwargs:
      self._children = {**self._children, **kwargs}
    for key in self._children.keys():
      self._children[key]._isChild = True
    return self
  def paginate(self):
    raise NotImplementedError('Child class should implement paginate method')
  def create(self):
    raise NotImplementedError('Child class should implement create method')
  def update(self):
    raise NotImplementedError('Child class should implement update method')
  def upsert(self):
    raise NotImplementedError('Child class should implement upsert method')
  def delete(self):
    raise NotImplementedError('Child class should implement delete method')
  def nested(self):
    raise NotImplementedError('Child class should implement nested method')
  def apply(self):
    raise NotImplementedError('Child class should implement apply method')
  def get(self):
    """Builds serialized query and returns data from GraphQL Endpoint
    
    Returns
    -------
    GraphQLType
      Instance of the GraphQLType class that created this QueryBuilder
    """
    _query = self.compile()
    data, errors = self._gql.query(_query)
    if self._type == QueryType.SINGULAR:
      if errors:
        instance = self._kind()
        instance.__errors__ = errors
        return instance
      instance = self._kind(**data)
      instance.__sync__ = True
      return instance
    else:
      # QueryType.PLURAL
      if errors:
        instances = [self._kind()]
        instances[0].__errors__ = errors
        return instances
      instances = [self._kind(**datum) for datum in data]
      # mark as synced (AKA: data ok)
      for i, instance in enumerate(instances):
        instances[i].__sync__ = True
      return instances
  def prettify_compile(self):
    _query = self.compile()
    tabs = 0
    pretty_query = ''
    for i, ch in enumerate(_query):
      # get context
      prev_ch = _query[i-1] if i > 0 else ''
      next_ch = _query[i+1] if i < len(_query) - 1 else ''
      tabs += 1 if ch in ['{', '('] else (-1 if next_ch in ['}', ')'] else 0)
      if ch != ' ' or prev_ch == ':':
        pretty_query += ch
        continue
      if next_ch in ['{', '(']:
        pretty_query += ch # add whitespace normally
      elif next_ch in ['}', ')']:
        pretty_query += f"\n{'  ' * tabs}"
      else:
        pretty_query += f"\n{'  ' * tabs}"
    return pretty_query
  def compile(self, defaultName=None, isChild=False) -> str:
    """Compiles collected parameters into a serialized query
    
    Parameters
    ----------
    isChild : bool, optional
      If true, looses the external query boilerplate ("query { ... }"), by default False
    
    Returns
    -------
    str
      Serialized query ("query Type { getType{fields...}}")
    """
    self._isChild = isChild
    if self._type == QueryType.PLURAL:
      return self.compile_plural(defaultName=defaultName, isChild=isChild)
    else:
      return self.compile_singular(defaultName=defaultName, isChild=isChild)
  def compile_plural(self, defaultName = None, isChild=False):
    # get query name
    if not defaultName:
      engine = inflect.engine()
      query_name = engine.plural(remove_capital(self._kind.__name__))
    else:
      query_name = defaultName
    _query = 'query { $Q_NAME $Q_PARAMS{ $Q_FIELDS } }' if not self._isChild else '$Q_NAME $Q_PARAMS{ $Q_FIELDS }'
    has_params = any([self._filter, self._orderBy, self._limit])
    _query = _query.replace('$Q_NAME', query_name)
    if has_params:
      _query = _query.replace('$Q_PARAMS', f'( {self.compile_plural_params()} )')
    else:
      _query = _query.replace('$Q_PARAMS', '')
    _query = _query.replace('$Q_FIELDS', self.compile_fields())
    return _query
  def compile_singular(self, defaultName = None, isChild=False):
    # get query name
    if not defaultName:
      query_name = remove_capital(self._kind.__name__)
    else:
      query_name = defaultName
    _query = 'query { $Q_NAME $Q_PARAMS{ $Q_FIELDS } }' if not self._isChild else '$Q_NAME $Q_PARAMS{ $Q_FIELDS }'
    has_params = any([self._findBy])
    _query = _query.replace('$Q_NAME', query_name)
    if has_params:
      _query = _query.replace('$Q_PARAMS', f'( {self.compile_singular_params()} )')
    else:
      _query = _query.replace('$Q_PARAMS', '')
    _query = _query.replace('$Q_FIELDS', self.compile_fields())
    return _query
  def compile_fields(self):
    compiled_child_queries = ''
    if self._children:
      child_queries = [child.compile(defaultName=qname, isChild=True) for qname, child in self._children.items()]
      compiled_child_queries = ' ' + ' '.join(child_queries)
    return " ".join(self._kind.get_base_fields()) + compiled_child_queries
  def compile_plural_params(self):
    # apply filter params
    filter_params = self.gql_param_dumps('filter', self._filter) if self._filter else ''
    # apply orderBy param
    orderby_params = self.gql_param_dumps('orderBy', self._orderBy, True) if self._orderBy else ''
    # apply limit
    limit_params = f'limit: {self._limit}' if self._limit else ''
    all_params = [filter_params, orderby_params, limit_params]
    used_params = [p for p in all_params if not p == '']
    return f' '.join(used_params)
  def compile_singular_params(self):
    if 'id' in self._findBy:
      k = 'id'
      v = self._findBy['id']
      findby_params = self.gql_param_dumps(k, v)
    else:
      findby_params = self.gql_param_dumps('findBy', self._findBy)
    return findby_params
  def gql_param_dumps(self, k, v, force_enum=False):
    """Converts GraphQL parameter pair (key, value) to it's serialized form
    Examples:
    - {'capacity': 10} -> 'capacity: 10'
    - {'enabled': True} -> 'enabled: true'
    - {'name': 'L101'} -> 'name: "L101"'
    - {'status': 'FINISHED'} -> 'status: FINISHED'
    - {
        'after': {
          'attribute': 'INSERTED_AT',
          'date': datetime(2020, 4, 4, 9, 12, 47, 1662)
        }
      } -> 'after: { attribute: INSERTED_AT date: "2020-04-04T09:12:47.1662Z" }'
    
    Parameters
    ----------
    k : param key
        left-hand side id for graphql parameter pair
    v : param value
        right-hand side value for graphql parameter pair
    """
    if type(v) == int:
      return f'{k}: {v}'
    if type(v) == float:
      return f'{k}: {v}'
    if type(v) == str:
      if (v.isupper() and not contains_digits(v) and len(v) > 3) or force_enum:
        # SCREAMING_CASE, this means it's an Enum, loose the quotes
        return f'{k}: {v}'
      else:
        # regular case, this means it's a string, include double quotes
        return f'{k}: "{v}"'
    if type(v) == bool:
      return f'{k}: {["false","true"][v]}' # short-hand for inline if
    if type(v) == datetime:
      utc_date = get_utc_date(v)
      return f'{k}: "{get_iso8601_str(utc_date)}"'
    if type(v) == dict:
      subparams = " ".join([self.gql_param_dumps(subk, subv, force_enum) for subk, subv in v.items()])
      return f'{k}: {{ {subparams} }}'
  # ! Method aliases:
  # filter
  fil = filter
  # filter sub-aliases:
  # filter(after={'attribute': 'INSERTED_AT', 'date': "2020-02-27T23:01:44Z"})
  def after(self, **kwargs):
    _kwargs = {}
    if 'attr' in kwargs:
      _kwargs['attribute'] = kwargs['attr']
      _kwargs['date'] = kwargs.get('date')
    else:
      _kwargs = {**kwargs}
    return self.filter(after=_kwargs)
  # filter(before={'attribute': 'INSERTED_AT', 'date': "2020-02-27T23:01:44Z"})
  def before(self, **kwargs):
    _kwargs = {}
    if 'attr' in kwargs:
      _kwargs['attribute'] = kwargs['attr']
      _kwargs['date'] = kwargs.get('date')
    else:
      _kwargs = {**kwargs}
    return self.filter(before=_kwargs)
  # filter(nullAttribute={ 'isNull': true 'attribute': 'DESCRIPTION' })
  def isNull(self, **kwargs):
    _kwargs = {}
    if 'attr' in kwargs:
      _kwargs['attribute'] = kwargs['attr']
    else:
      _kwargs['attribute'] = kwargs.get('attribute')
    if 'value' in kwargs:
      _kwargs['isNull'] = kwargs['value']
    else:
      _kwargs['isNull'] = kwargs.get('isNull', True) # Default to True
    return self.filter(nullAttribute=_kwargs)
  # findBy
  fb = findBy
  find = findBy
  # orderBy
  ob = orderBy
  order = orderBy
  # limit
  lim = limit
  # offset
  off = offset
  # children
  child = children
  ch = children
  # get
  run = get
  # paginate
  pgt = paginate
  # create
  # update
  # upsert
  # delete
  # nested
  # apply




def main():
  print('Hello, GSTORM!')
  # ! TODO: create base class for Gstorm, other classes should inherit from it
  gql = setup_gql('dev')
  # * 1: Get data:
  GET_TANKS = '{ tanks{ id name capacity } }'
  tanks_data, errors = gql.query(GET_TANKS)
  if errors:
    pprint(errors)
    return
  tanks = [Tank(**tank) for tank in tanks_data]
  # * 2: Process data:
  pprint(f'Tank count: {len(tanks)}')
  empty_tanks = [tank for tank in tanks if tank.capacity == 0]
  print('Empty tanks:')
  pprint(len(empty_tanks))
  empty_tanks[10].capacity = 1234
  pprint(empty_tanks)
  # * re-upload data:
  UPDATE_TANK = '''
  mutation UPDATE_TANK($id: Int! $tank: UpdateTankParams!){
    updateTank(
      id: $id
      tank: $tank
    ){
      successful
      messages{
        field
        message
      }
      result{
        id
        name
        capacity
      }
    }
  }
  '''
  payload, errors = gql.mutate( mutation = UPDATE_TANK, variables= {
    'id': int(empty_tanks[10].id),
    'tank': {
      'name': empty_tanks[10].name,
      'capacity': empty_tanks[10].capacity
    }
  })
  if errors:
    print(errors)
  # * STORM
  Tank = storm.getType('Tank')
  tanks = Tank.get(filter={'capacity': 0})
  tanks[-1].capacity = 1234
  tanks[-1].update()
  storm.batchUpdate([GraphqlType])
  tank = Tank()
  tank.
  for tank_batch in storm.paginate(Tank):
    for tank in tank_batch:
      


if __name__ == "__main__":
  main()


one_to_one_map = {
    "String": ("str", "''"),
    "Int": ("int", "0"),
    "Float": ("float", "0.0"),
    "Date": ("date", "date.today()") ,
    "Time": ("time", "datetime.now().time()") ,
    "DateTime": ("datetime", "datetime.now()"),
    "Boolean": ("bool", "False"),
    "map": ("dict", "{}"),
    "Text": ("str", "''"),
    "Enum": ("str", "''")
}

imports = """
from typing import List, Union
from datetime import datetime, date, time
from gstorm.GraphQLType import GraphQLType
from dataclasses import dataclass, field
from gstorm.helpers.typing_helpers import convert_to, list_convert_to, gql_repr, gql_list_repr
import attr
"""

example = {'Author': {'fields': [{'has_one': False,
                        'name': 'name',
                        'type': {'kind': 'SCALAR', 'name': 'String'},
                        'unique': True},
                       {'has_one': False,
                        'name': 'lastName',
                        'type': {'kind': 'SCALAR', 'name': 'String'},
                        'unique': True},
                       {'has_one': False,
                        'name': 'finishTurn',
                        'type': {'kind': 'SCALAR', 'name': 'Time'},
                        'unique': False},
                       {'has_one': False,
                        'name': 'dateOfBirth',
                        'type': {'kind': 'SCALAR', 'name': 'Date'},
                        'unique': False},
                       {'has_one': False,
                        'name': 'active',
                        'type': {'kind': 'SCALAR', 'name': 'Boolean'},
                        'unique': False},
                       {'has_one': False,
                        'name': 'blogPosts',
                        'type': {'kind': 'LIST', 'name': 'BlogPost'},
                        'unique': False},
                       {'has_one': False,
                        'name': 'comments',
                        'type': {'kind': 'LIST', 'name': 'Comment'},
                        'unique': False},
                       {'has_one': True,
                        'name': 'avatar',
                        'type': {'kind': 'OBJECT', 'name': 'Avatar'},
                        'unique': False},
                       {'has_one': False,
                        'name': 'categories',
                        'type': {'kind': 'LIST', 'name': 'Category'},
                        'unique': False},
                       {'has_one': False,
                        'name': 'session',
                        'type': {'kind': 'SCALAR', 'name': 'Float'},
                        'unique': False},
                       {'has_one': False,
                        'name': 'email',
                        'type': {'kind': 'SCALAR', 'name': 'String'},
                        'unique': False}],
            'kind': 'TYPE',
            'name': 'Author'},
 'Avatar': {'fields': [{'has_one': False,
                        'name': 'imageUrl',
                        'type': {'kind': 'SCALAR', 'name': 'String'},
                        'unique': True},
                       {'has_one': False,
                        'name': 'avatarConfigurations',
                        'type': {'kind': 'LIST', 'name': 'AvatarConfiguration'},
                        'unique': False},
                       {'has_one': False,
                        'name': 'comments',
                        'type': {'kind': 'LIST', 'name': 'Comment'},
                        'unique': False},
                       {'has_one': False,
                        'name': 'owner',
                        'type': {'kind': 'OBJECT', 'name': 'Author'},
                        'unique': False},
                       {'has_one': False,
                        'name': 'blogPost',
                        'type': {'kind': 'OBJECT', 'name': 'BlogPost'},
                        'unique': False},
                       {'has_one': False,
                        'name': 'providers',
                        'type': {'kind': 'LIST', 'name': 'Reply'},
                        'unique': False},
                       {'has_one': False,
                        'name': 'consumers',
                        'type': {'kind': 'LIST', 'name': 'Reply'},
                        'unique': False}],
            'kind': 'TYPE',
            'name': 'Avatar'},
 'AvatarConfiguration': {'fields': [{'has_one': False,
                                     'name': 'guid',
                                     'type': {'kind': 'SCALAR',
                                              'name': 'String'},
                                     'unique': True},
                                    {'has_one': False,
                                     'name': 'config',
                                     'type': {'kind': 'OBJECT',
                                              'name': 'Avatar'},
                                     'unique': False},
                                    {'has_one': False,
                                     'name': 'version',
                                     'type': {'kind': 'SCALAR',
                                              'name': 'String'},
                                     'unique': False}],
                         'kind': 'TYPE',
                         'name': 'AvatarConfiguration'},
 'BlogPost': {'fields': [{'has_one': False,
                          'name': 'author',
                          'type': {'kind': 'OBJECT', 'name': 'Author'},
                          'unique': False},
                         {'has_one': False,
                          'name': 'name',
                          'type': {'kind': 'SCALAR', 'name': 'String'},
                          'unique': True},
                         {'has_one': False,
                          'name': 'body',
                          'type': {'kind': 'SCALAR', 'name': 'Text'},
                          'unique': False},
                         {'has_one': False,
                          'name': 'title',
                          'type': {'kind': 'SCALAR', 'name': 'String'},
                          'unique': False},
                         {'has_one': False,
                          'name': 'lastName',
                          'type': {'kind': 'SCALAR', 'name': 'String'},
                          'unique': False},
                         {'has_one': False,
                          'name': 'dateOfPublication',
                          'type': {'kind': 'SCALAR', 'name': 'Date'},
                          'unique': False},
                         {'has_one': False,
                          'name': 'statusWord',
                          'type': {'kind': 'ENUM', 'name': 'StatusWord'},
                          'unique': False},
                         {'has_one': False,
                          'name': 'comments',
                          'type': {'kind': 'LIST', 'name': 'Comment'},
                          'unique': False},
                         {'has_one': False,
                          'name': 'approvedAt',
                          'type': {'kind': 'SCALAR', 'name': 'DateTime'},
                          'unique': False},
                         {'has_one': True,
                          'name': 'alias',
                          'type': {'kind': 'OBJECT', 'name': 'Avatar'},
                          'unique': False},
                         {'has_one': False,
                          'name': 'tag',
                          'type': {'kind': 'ENUM', 'name': 'Tag'},
                          'unique': False},
                         {'has_one': True,
                          'name': 'wordCount',
                          'type': {'kind': 'OBJECT', 'name': 'WordCount'},
                          'unique': False}],
              'kind': 'TYPE',
              'name': 'BlogPost'},
 'Category': {'fields': [{'has_one': False,
                          'name': 'name',
                          'type': {'kind': 'SCALAR', 'name': 'String'},
                          'unique': True},
                         {'has_one': False,
                          'name': 'description',
                          'type': {'kind': 'SCALAR', 'name': 'String'},
                          'unique': False},
                         {'has_one': False,
                          'name': 'createdBy',
                          'type': {'kind': 'OBJECT', 'name': 'Author'},
                          'unique': False},
                         {'has_one': False,
                          'name': 'editedBy',
                          'type': {'kind': 'OBJECT', 'name': 'Author'},
                          'unique': False}],
              'kind': 'TYPE',
              'name': 'Category'},
 'Comment': {'fields': [{'has_one': False,
                         'name': 'author',
                         'type': {'kind': 'OBJECT', 'name': 'Author'},
                         'unique': False},
                        {'has_one': False,
                         'name': 'supremeAuthor',
                         'type': {'kind': 'OBJECT', 'name': 'Author'},
                         'unique': False},
                        {'has_one': False,
                         'name': 'blogPost',
                         'type': {'kind': 'OBJECT', 'name': 'BlogPost'},
                         'unique': False},
                        {'has_one': False,
                         'name': 'body',
                         'type': {'kind': 'SCALAR', 'name': 'String'},
                         'unique': False},
                        {'has_one': False,
                         'name': 'recievedAt',
                         'type': {'kind': 'SCALAR', 'name': 'DateTime'},
                         'unique': False},
                        {'has_one': False,
                         'name': 'reportedAt',
                         'type': {'kind': 'SCALAR', 'name': 'DateTime'},
                         'unique': False},
                        {'has_one': False,
                         'name': 'avatar',
                         'type': {'kind': 'OBJECT', 'name': 'Avatar'},
                         'unique': False},
                        {'has_one': False,
                         'name': 'styles',
                         'type': {'kind': 'LIST', 'name': 'Style'},
                         'unique': False},
                        {'has_one': False,
                         'name': 'finishedAt',
                         'type': {'kind': 'SCALAR', 'name': 'Int'},
                         'unique': False},
                        {'has_one': False,
                         'name': 'type',
                         'type': {'kind': 'SCALAR', 'name': 'Int'},
                         'unique': False},
                        {'has_one': False,
                         'name': 'tag',
                         'type': {'kind': 'ENUM', 'name': 'Tag'},
                         'unique': False}],
             'kind': 'TYPE',
             'name': 'Comment'},
 'Employee': {'fields': [{'has_one': False,
                          'name': 'name',
                          'type': {'kind': 'SCALAR', 'name': 'String'},
                          'unique': False},
                         {'has_one': False,
                          'name': 'email',
                          'type': {'kind': 'SCALAR', 'name': 'String'},
                          'unique': True},
                         {'has_one': False,
                          'name': 'department',
                          'type': {'kind': 'SCALAR', 'name': 'String'},
                          'unique': False},
                         {'has_one': False,
                          'name': 'employer',
                          'type': {'kind': 'OBJECT', 'name': 'Employee'},
                          'unique': False},
                         {'has_one': False,
                          'name': 'employees',
                          'type': {'kind': 'LIST', 'name': 'Employee'},
                          'unique': False},
                         {'has_one': False,
                          'name': 'boss',
                          'type': {'kind': 'OBJECT', 'name': 'Employee'},
                          'unique': False},
                         {'has_one': False,
                          'name': 'workers',
                          'type': {'kind': 'LIST', 'name': 'Employee'},
                          'unique': False}],
              'kind': 'TYPE',
              'name': 'Employee'},
 'Equipment': {'fields': [{'has_one': False,
                           'name': 'name',
                           'type': {'kind': 'SCALAR', 'name': 'String'},
                           'unique': True},
                          {'has_one': False,
                           'name': 'description',
                           'type': {'kind': 'SCALAR', 'name': 'Text'},
                           'unique': False},
                          {'has_one': False,
                           'name': 'information',
                           'type': {'kind': 'SCALAR', 'name': 'map'},
                           'unique': False}],
               'kind': 'TYPE',
               'name': 'Equipment'},
 'OrderComment': {'fields': [{'has_one': False,
                              'name': 'description',
                              'type': {'kind': 'SCALAR', 'name': 'Text'},
                              'unique': False}],
                  'kind': 'TYPE',
                  'name': 'OrderComment'},
 'Reply': {'fields': [{'has_one': False,
                       'name': 'title',
                       'type': {'kind': 'SCALAR', 'name': 'String'},
                       'unique': True},
                      {'has_one': False,
                       'name': 'provider',
                       'type': {'kind': 'OBJECT', 'name': 'Avatar'},
                       'unique': False},
                      {'has_one': False,
                       'name': 'consumer',
                       'type': {'kind': 'OBJECT', 'name': 'Avatar'},
                       'unique': False},
                      {'has_one': False,
                       'name': 'struct',
                       'type': {'kind': 'SCALAR', 'name': 'map'},
                       'unique': False}],
           'kind': 'TYPE',
           'name': 'Reply'},
 'StatusWord': {'enumValues': ['APPROVED', 'PENDING', 'REJECTED'],
                'kind': 'ENUM',
                'name': 'StatusWord'},
 'Style': {'fields': [{'has_one': False,
                       'name': 'description',
                       'type': {'kind': 'SCALAR', 'name': 'Text'},
                       'unique': False},
                      {'has_one': False,
                       'name': 'comment',
                       'type': {'kind': 'OBJECT', 'name': 'Comment'},
                       'unique': False}],
           'kind': 'TYPE',
           'name': 'Style'},
 'Tag': {'enumValues': ['QUESTION', 'DEPENDENCY', 'NEWS'],
         'kind': 'ENUM',
         'name': 'Tag'},
 'WordCount': {'fields': [{'has_one': False,
                           'name': 'outputBlogPost',
                           'type': {'kind': 'OBJECT', 'name': 'BlogPost'},
                           'unique': False},
                          {'has_one': False,
                           'name': 'count',
                           'type': {'kind': 'SCALAR', 'name': 'Int'},
                           'unique': False}],
               'kind': 'TYPE',
               'name': 'WordCount'}}
